import os
import pandas
import numpy as np
import warnings
from . import io


__all__ =["get_target_lightcurve"]



ZTFCOLOR = { # ZTF
        "p48r":dict(marker="o",ms=7,  mfc="C3"),
        "p48g":dict(marker="o",ms=7,  mfc="C2"),
        "p48i":dict(marker="o",ms=7, mfc="C1")
}

BAD_ZTFCOLOR = { # ZTF
        "p48r":dict(marker="o",ms=6,  mfc="None", mec="C3"),
        "p48g":dict(marker="o",ms=6,  mfc="None", mec="C2"),
        "p48i":dict(marker="o",ms=6,  mfc="None", mec="C1")
}


def get_target_lightcurve(name, load_salt2param=True, **kwargs):
    """ load the lightcurve object for the given target. """
    return LightCurve.from_name(name, load_salt2param=load_salt2param, **kwargs)
    
# ================== #
#                    #
#   LIGHTCURVES      #
#                    #
# ================== #
    
class LightCurve( object ):

    def __init__(self, data, meta=None, salt2param=None, use_dask=False):
        """ likely, this is not how you should load the data. 
        See from_name() or from_filename() class methods.
        """
        self.set_data(data)
        self.set_meta(meta)
        self.set_salt2param(salt2param)
        self._use_dask = use_dask
        
    @classmethod
    def from_filename(cls, filename, use_dask=False):
        """  load a Lightcurve object from a given file. """
        if use_dask:
            from dask import delayed
            # This is faster than dd.read_cvs and does what we need here
            lc = delayed(pandas.read_csv)(filename,  delim_whitespace=True, comment='#')
        else:
            lc = pandas.read_csv(filename,  delim_whitespace=True, comment='#')
            
        meta = pandas.Series([os.path.basename(filename).split("_")[0]], 
                             index=["name"])
        return cls(lc, meta=meta, use_dask=use_dask)

    @classmethod
    def from_name(cls, targetname, use_dask=False, load_salt2param=True, **kwargs):
        """ """
        filename = io.get_target_lc(targetname)
        this = cls.from_filename(filename, use_dask=use_dask, **kwargs)
        this._targetname = targetname
        if load_salt2param:
            this.load_salt2param()
            
        return this
    # ================ #
    #    Method        #
    # ================ #
    # --------- #
    #  LOADER   #
    # --------- #
    def load_salt2param(self):
        """ """
        targetname = self.targetname
        if targetname is None:
            warnings.warn("Unknown targetname (=None) ; use manually set_salt2param()")
            return None
        
        from .salt2 import get_target_salt2param
        salt2param = get_target_salt2param(targetname)
        self.set_salt2param(salt2param)
        
    # --------- #
    #  SETTER   #
    # --------- #
    def set_data(self, data):
        """ """
        self._data = data

    def set_salt2param(self, salt2param):
        """ """
        self._salt2param = salt2param
        
    def set_meta(self, meta):
        """ """
        self._meta = meta  

    # --------- #
    #  GETTER   #
    # --------- #
    def get_obsphase(self, min_detection=5, groupby=None, **kwargs):
        """ 
        Returns
        -------
        pandas.Series 
        """
        lcdata = self.get_lcdata(min_detection=min_detection, **kwargs)
        if groupby is None:
            return lcdata["phase"]
        
        return lcdata.groupby(groupby)["phase"].apply( list )
        
    def get_saltmodel(self):
        """ """
        from .salt2 import get_saltmodel
        return get_saltmodel(**self.salt2param.rename({"redshift":"z"}
                                                     )[["z","t0","x0","x1","c",
                                                        "mwebv"]].to_dict()
                            )
    def get_lcdata(self, zp=None, in_mjdrange=None,
                       min_detection=None,
                       filters=None,
                       flagout=[1,2,4,8,16]):
        """ 
        filters: [string, None or list]
            list of filters 
            - None/'*' or 'all': no filter selection/
            - string: just this filter (e.g. 'p48g')
            - list of string: just these filters (e.g. ['p48g','p48r'])

        flagout: [list of int or string]
            flag == 0 means all good, but in details:
            
            0: no warning 
            1: flux_err==0 Remove unphysical errors 
            2: chi2dof>3: Remove extreme outliers 
            4: cloudy>1: BTS cut 
            8: infobits>0: BTS cut 16: mag_lim<19.3: Cut applied in Dhawan 2021 
            32: seeing>3: Cut applied in Dhawan 2021 
            64: fieldid>879: Recommended IPAC cut 
            128: moonilf>0.5: Recommended IPAC cut 
            256: has_baseline>1: Has a valid baseline correction 
            512: airmass>2: Recommended IPAC cut 
            1024: flux/flux_err>=5: Nominal detection

        """
        from .utils import flux_to_mag

        if flagout in ["all","any","*"]:
            data = self.data[self.data["flag"]==0]
            
        if flagout in ["all","any","*"]:
            data = self.data[self.data["flag"]==0]
        elif flagout is None:
            data = self.data.copy()
        else:
            flag_ = np.all([(self.data.flag&i_==0) for i_ in np.atleast_1d(flagout)], axis=0)
            data = self.data[flag_]            
            

        if zp is None:
            zp = data["ZP"].values
            coef = 1. 
        else:
            coef = 10 ** (-(data["ZP"].values - zp) / 2.5)

            
        flux  = data["flux"] * coef
        error = data["flux_err"] * coef
        detection = flux/error
            
        
        lcdata = data[["mjd","mag","mag_err","filter","field_id","x_pos","y_pos", "flag"]]
        additional = pandas.DataFrame(np.asarray([zp, flux, error, detection]).T,
                                         columns=["zp", "flux", "error", "detection"],
                                         index=lcdata.index)
        
        additional["mag_lim"], _ = flux_to_mag(error*5, None, zp=zp)
        
        lcdata = pandas.merge(lcdata, additional, left_index=True, right_index=True)
#        lcdata.loc["zp",:] = zp
#        lcdata["flux"] = flux
#        lcdata["error"] = error
#        lcdata["detection"] = detection
        lcdata["filter"] = lcdata["filter"].replace("ztfg","p48g").replace("ztfr","p48r").replace("ztfi","p48i") 
        
        
        if self.has_salt2param():
            lcdata["phase"] = lcdata["mjd"]-self.salt2param['t0']
        else:
            lcdata["phase"] = np.NaN
            
        if in_mjdrange is not None:
            lcdata = lcdata[lcdata["mjd"].between(*in_mjdrange)]

        if min_detection is not None:
            lcdata = lcdata[lcdata["detection"]>min_detection]

        if filters is not None and filters not in ["*","all"]:
            lcdata = lcdata[lcdata["filter"].isin(np.atleast_1d(filters))]
            
        return lcdata


    def get_model_residual(self, model=None, modelprop={}, 
                           intrinsic_error=None,
                           **kwargs):
        """ get a dataframe with lightcurve data, model and residuals information.

        Parameters
        ----------
        model: [sncosmo.Model or None] -optional-
            provide the sncosmo model from which the model flux can be obtained.
            If None given [default] this method will call self.get_saltmodel().

        modelprop: [dict] -optional-
            kwarg information passed to model.set() to change the default model parameters.
            = This is only used in model=None = 
            It aims at updating model given by self.get_saltmodel()
            
        intrinsic_error: [float] -optional-
            provide an intrinsic error for the lightcurve. This will be stored as 'error_int'
            in the returned DataFrame. 
            The 'pull' will use the quadratic sum of error and error_int to be calculated.
            if None given [default] this will assume 0.

        
        **kwargs goes to get_lcdata()

        Returns
        -------
        DataFrame
        """
        basedata = self.get_lcdata(**kwargs)[["mjd","flux", "error","phase", "filter","flag"]]
        if model is None:
            model = self.get_saltmodel()
            if modelprop is not None:
                # Does nothing if empty dict
                model.set(**modelprop) 

        # Model
        basedata["model"] = model.bandflux(basedata["filter"], basedata["mjd"], 
                                            zp=self.flux_zp, zpsys="ab")
        # Residual    
        basedata["residual"] = basedata["flux"] - basedata["model"]

        # Error    
        basedata["error_int"] = intrinsic_error if intrinsic_error is not None else 0
        total_error = np.sqrt(basedata["error"]**2 + basedata["error_int"]**2)
        # Pull    
        basedata["pull"] = basedata["residual"]/total_error
        return basedata

    
    def get_sncosmotable(self, min_detection=5,  phase_range=[-10,30], filters=["p48r","p48g"], **kwargs):
        """ """
        from .utils import mag_to_flux
        
        t0 = self.salt2param["t0"]
        to_fit = self.get_lcdata(min_detection=min_detection, in_mjdrange= t0 + phase_range,
                                 filters=filters, **kwargs)
        sncosmo_lc = to_fit.rename({"mjd":"time", "filter":"band"}, axis=1)[["time","band","zp","mag","mag_err"]]
        sncosmo_lc["zpsys"] = "ab"
        sncosmo_lc["flux"], sncosmo_lc["flux_err"] = mag_to_flux(sncosmo_lc["mag"],
                                                                 sncosmo_lc["mag_err"],
                                                                 zp=sncosmo_lc["zp"])
        return sncosmo_lc

    def fit_salt(self, free_parameters=['t0', 'x0', 'x1', 'c'],
                       min_detection=5, phase_range=[-10,30], filters=["p48r","p48g"],
                       as_dataframe=False,
                       **kwargs):
        """ """
        import sncosmo
        from astropy import table
        from .salt2 import salt2result_to_dataframe
        model = self.get_saltmodel()
        sncosmo_df = self.get_sncosmotable(min_detection=min_detection, 
                                           phase_range=phase_range, 
                                           filters=filters)
        fitted_data = table.Table.from_pandas(sncosmo_df)
        (result, fitted_model) = sncosmo.fit_lc(fitted_data, model,
                                            vparam_names=free_parameters,  **kwargs)
        if as_dataframe:
            return salt2result_to_dataframe(result)
        
        return (fitted_data,model), (result, fitted_model)

    def fit_salt_perfilter(lc, filters=["p48g",'p48r','p48i'],
                           min_detection=5, free_parameters=['t0','x0', 'x1'],
                           phase_range=[-10, 30], t0_range=[-2,+2]):
        """ """
        results = []
        for filter_ in filters:
            try:
                result  = lc.fit_salt(min_detection=min_detection,
                                      filters=filter_,
                                      free_parameters=free_parameters,
                                      phase_range=phase_range,
                                      bounds={"t0":lc.salt2param['t0']+t0_range},
                                      as_dataframe=True
                                      )
            except:
                warnings.warn(f"failed for filter {filter_}")
                result = pandas.DataFrame()

            results.append(result)
            
        return pandas.concat(results, keys=filters)
        
    
    

    # --------- #
    #  PLOTTER  #
    # --------- #
    def show(self, ax=None, figsize=None, zp=None, formattime=True, zeroline=True,
                 incl_salt2=True, autoscale_salt=True, clear_yticks=True,
                 phase_range=[-30,100],
                 zprop={}, inmag=False, ulength=0.1, ualpha=0.1, notplt=False, **kwargs):
        """ """
        from matplotlib import dates as mdates
        from astropy.time import Time
        
        self._compute()
        # - Axes Definition
        if ax is None:
            import matplotlib.pyplot as mpl
            fig = mpl.figure(figsize=[7,4])# if figsize is None else figsize)
            ax = fig.add_axes([0.1,0.15,0.8,0.75])
        else:
            fig = ax.figure
            
        # - End axes definition
        # -- 
        # - Data
        base_prop = dict(ls="None", mec="0.9", mew=0.5, ecolor="0.7", zorder=7)
        bad_prop  = dict(ls="None", mew=1, ecolor="0.7", zorder=6)        
        lineprop  = dict(color="0.7", zorder=1, lw=0.5)

        if incl_salt2:
            saltmodel = self.get_saltmodel()
        else:
             saltmodel = None
             autoscale_salt = False
             
        modeltime = self.salt2param.t0 + np.linspace(-15,50,100)
        t0 = self.salt2param.t0
        if phase_range is not None:
            timerange = [t0-30, t0+100]
        else:
            timerange = None
            
        lightcurves = self.get_lcdata(zp=zp, in_mjdrange=timerange)        
        bands = np.unique(lightcurves["filter"])
        
        # flag goods
        flag_good = (lightcurves.flag&1==0) & (lightcurves.flag&2==0) & (lightcurves.flag&4==0) & (lightcurves.flag&8==0)

        max_saltlc = 0
        min_saltlc = 100
        # Loop over bands
        for band_ in bands:
            if band_ not in ZTFCOLOR:
                warnings.warn(f"WARNING: Unknown instrument: {band_} | magnitude not shown")
                continue

            flagband   = (lightcurves["filter"]==band_)
            
            bdata = lightcurves[flagband]
            flag_good_ = flag_good[flagband]
            
            # IN FLUX
            if not inmag:
                # - Data
                datatime = Time(bdata["mjd"], format="mjd").datetime
                y, dy = bdata["flux"], bdata["error"]
                # - Salt
                if saltmodel is not None:
                    saltdata = saltmodel.bandflux(band_, modeltime, zp=self.flux_zp, zpsys="ab") if saltmodel is not None else None
                else:
                    saltdata = None
                    
            # IN MAG                                
            else:
                flag_det = (lightcurves["mag"]<99)
                # - Data                
                bdata = bdata[flag_det]
                flag_good_ = flag_good_[flag_det]
                datatime = Time(bdata["mjd"], format="mjd").datetime
                y, dy = bdata["mag"], bdata["mag_err"]
                # - Salt
                if saltmodel is not None:
                    saltdata = saltmodel.bandmag(band_, "ab",modeltime) if saltmodel is not None else None
                else:
                    saltdata = None
            
            # -> good
            ax.errorbar(datatime[flag_good_],
                            y[flag_good_],  yerr=dy[flag_good_], 
                            label=band_, 
                            **{**base_prop, **ZTFCOLOR[band_],**kwargs}
                            )
            # -> bad
            ax.errorbar(datatime[~flag_good_],
                            y[~flag_good_],  yerr=dy[~flag_good_], 
                            label=band_, 
                            **{**bad_prop, **BAD_ZTFCOLOR[band_],**kwargs}
                            )
        
            if saltdata is not None:
                ax.plot(Time(modeltime, format="mjd").datetime,
                        saltdata,
                        color=ZTFCOLOR[band_]["mfc"], zorder=5)

                max_saltlc = np.max([max_saltlc, np.max(saltdata)])
                min_saltlc = np.min([min_saltlc, np.min(saltdata)])
            
        if inmag:
            ax.invert_yaxis()
            for band_ in bands:
                bdata = lightcurves[(lightcurves["filter"]==band_) & (lightcurves["mag"]>=99)]
                datatime = Time(bdata["mjd"], format="mjd").datetime
                y = bdata["mag_lim"]
                ax.errorbar(datatime, y,
                                 yerr=ulength, lolims=True, alpha=ualpha,
                                 color=ZTFCOLOR[band_]["mfc"], 
                                 ls="None",  label="_no_legend_")
                                 
                                 
                
        if formattime:
            locator = mdates.AutoDateLocator()
            formatter = mdates.ConciseDateFormatter(locator)
            ax.xaxis.set_major_locator(locator)
            ax.xaxis.set_major_formatter(formatter)

        lunit = "flux" if not inmag else "mag"
        ax.set_ylabel(f"{lunit} [zp={zp}]" if zp is not None else f"{lunit} []")
        if zeroline:
            ax.axhline(0 if not inmag else 22, **{**dict(color="0.7",ls="--",lw=1, zorder=1),**zprop} )

        if not inmag:
            max_data = np.percentile(lightcurves["flux"], 99.)
            mean_error = np.nanmean(lightcurves["error"])
            ax.set_ylim(-2*mean_error, max_data*1.15)
            if clear_yticks:
                ax.axes.yaxis.set_ticklabels([])
            
        if autoscale_salt:
            if timerange is not None:
                ax.set_xlim(*Time(timerange,format="mjd").datetime)
            if not inmag:
                ax.set_ylim(bottom=-max_saltlc*0.25)
                ax.set_ylim(top=max_saltlc*1.25)
            else:
                if np.isinf(min_saltlc) or np.isnan(min_saltlc):
                    ax.set_ylim(23, 14)
                else:
                    ax.set_ylim(top=min_saltlc*0.95)
                    
        return fig
    
    # ----------- #
    #  Internal   #
    # ----------- #
    def _compute(self, persist=False):
        """ """
        if not self._use_dask:
            return 
        
        from dask.dataframe.core import DataFrame as daskDataFrame
        from dask.delayed import Delayed
        if not type(self._data) in [daskDataFrame, Delayed]:
            return
        if persist:
            self._data = self._data.persist()
        else:
            self._data = self._data.compute()
    
    # ================ #
    #   Properties     #
    # ================ #        
    # Baseline    
    @property
    def data(self):
        """ """
        return self._data
    
    @property
    def meta(self):
        """ """
        return self._meta

    @property
    def targetname(self):
        """ """
        if not hasattr(self, "_targetname"):
            if self.meta is not None:
                return self.meta.get("name",None)
            
            return None
        
        return self._targetname
    

    def has_salt2param(self):
        """" """
        return self.salt2param is not None
    
    @property
    def salt2param(self):
        """ """
        if not hasattr(self,"_salt2param"):
            return None
        return self._salt2param


    @property
    def flux_zp(self):
        """ """
        zp = self.data["ZP"]
        if len(np.unique(zp)) == 1:
            return float(zp[0])
        return zp
    
class LightCurveCollection( object ):
    
    def __init__(self, lightcurves, use_dask=False):
        """ """
        self.set_lightcurves(lightcurves)
        self._use_dask = use_dask
            
    @classmethod
    def from_directory(cls, directory, contains=None, startswith=None, 
                       extension=".csv", use_dask=False):
        """ """
        from glob import glob
        glob_format = "*" if not startswith else f"{startswith}*"
        if contains is not None:
            glob_format += f"{contains}*"
        if extension is not None:
            glob_format +=f"{extension}"
            
        files = glob(os.path.join(directory, glob_format))
        
        return cls.from_filenames(files, use_dask=use_dask)
    
    @classmethod
    def from_filenames(cls, filenames, use_dask=False, **kwargs):
        """ """
        return cls([LightCurve.from_filename(file_, use_dask=use_dask, **kwargs) 
                    for file_ in filenames], use_dask=use_dask)
        
    # ---------------- #
    #   Internal       #
    # ---------------- #
    def _persist(self):
        """ """
        if not self._use_dask:
            return 
        
        _ = [lc._compute(persist=True) for lc in self.lightcurves]
        
    # ================ #
    #    Method        #
    # ================ #
    def set_lightcurves(self, lightcurves):
        """ """
        self._lightcurves = lightcurves
    
    def get_meta(self, rebuild=False):
        """ """
        if rebuild:
            metas = self.call_down("meta", False)
            return pandas.DataFrame(metas)

        return self.meta

    def get_target_lightcurve(self, name):
        """ """
        lc_ = np.asarray(self.lightcurves)[self.meta.query(f"name == '{name}'").index]
        if len(lc_)==1:
            return lc_[0]

        return lc_

    def show_target(self, name, ax=None, **kwargs):
        """ """
        lctarget = self.get_target_lightcurve(name)
        return lctarget.show(ax=ax, **kwargs)
    
    # -------- #
    # INTERNAL #
    # -------- #
    def map_down(self, what, margs, *args, **kwargs):
        """ """
        return [getattr(spec, what)(marg, *args, **kwargs)
                for lc, marg in zip(self.lightcurves, margs)]
    
    def call_down(self, what, isfunc, *args, **kwargs):
        """ """
        if isfunc:
            return [getattr(lc,what)(*args, **kwargs) 
                    for lc in self.lightcurves]
        return [getattr(lc,what) for lc in self.lightcurves]    
    
    
    @property
    def meta(self):
        """ """
        if not hasattr(self,"_meta"):
            self._meta = self.get_meta(rebuild=True)
        return self._meta
    
    # Baseline    
    @property
    def lightcurves(self):
        """ """    
        return self._lightcurves



import numpy as np

def mag_to_flux(mag, magerr=None, units="zp", zp=25.0, wavelength=None):
    """converts magnitude into flux
    Parameters
    ----------
    mag: [float or array]
        AB magnitude(s)
    magerr: [float or array] -optional-
        magnitude error if any
    units: [string] -optional-
        Unit system in which to return the flux:
        - 'zp':   units base on zero point as required for sncosmo fits
        - 'phys': physical units [erg/s/cm^2/A)
    zp: [float or array] -optional-
        zero point of for flux; required if units == 'zp'
    wavelength: [float or array] -optional-
        central wavelength of the photometric filter.
        In Angstrom; required if units == 'phys'
    Returns
    -------
    - float or array (if magerr is None)
    - float or array, float or array (if magerr provided)
    """
    if units not in ["zp", "phys"]:
        raise ValueError("units must be 'zp' or 'phys'")
    elif units == "zp":
        if zp is None:
            raise ValueError("zp must be float or array if units == 'zp'")
        flux = 10 ** (-(mag - zp) / 2.5)
    else:
        if wavelength is None:
            raise ValueError("wavelength must be float or array if units == 'phys'")
        flux = 10 ** (-(mag + 2.406) / 2.5) / wavelength ** 2

    if magerr is None:
        return flux

    dflux = np.abs(flux * (-magerr / 2.5 * np.log(10)))  # df/f = dcount/count
    return flux, dflux


def flux_to_mag(flux, dflux, wavelength=None, zp=None, inhz=False):
    """ Converts fluxes (erg/s/cm2/A) into AB or zp magnitudes


    Parameters
    ----------
    flux, fluxerr: [float or array]
        flux and its error 

    wavelength: [float or array] -optional-
        = Ignored if inhz=True =
        central wavelength [in AA] of the photometric filter.


    zp: [float or array] -optional-
        = Ignored if inhz=True =
        = Ignored if wavelength is provided =
        zero point of for flux;

    inhz:
        set to true if the flux (and flux) are given in erg/s/cm2/Hz
        (False means in erg/s/cm2/AA)
        
    Returns
    -------
    - float or array (if magerr is None)
    - float or array, float or array (if magerr provided)
    
    """
    if inhz:
        zp = -48.598 # instaad of -48.60 such that hz to aa is correct
        wavelength = 1
    else:
        if zp is None and wavelength is None:
            raise ValueError("zp or wavelength must be provided")
        if zp is None:
            zp = -2.406 
        else:
            wavelength=1
            
    mag_ab = -2.5*np.log10(flux*wavelength**2) + zp
    if dflux is None:
        return mag_ab, None
    
    dmag_ab = +2.5/np.log(10) * dflux / flux
    return mag_ab, dmag_ab



def hist_colorbar(data, ax=None, vmin=None, vmax=None, bins="auto", 
                     fcolorbar=0.05, cmap="cividis", alpha=None,
                      clear_axes=True):
    """ """
    if vmin is None:
        vmin = np.nanmin(data)
    if vmax is None:
        vmax = np.nanmax(data)
        
    import matplotlib.pyplot as plt
    from matplotlib.colors import Normalize

    cmap = plt.colormaps[cmap]

    intensity, binegdes = np.histogram(data, range=[vmin, vmax], bins=bins)
    bins = {"edge":binegdes,
            "centroid":np.mean([binegdes[1:],binegdes[:-1]], axis=0),
            "width":binegdes[1:]-binegdes[:-1]}

    if ax is None:
        fig = plt.figure(figsize=[6,5])
        ax = fig.add_axes([0.1,0.15,0.8,0.75])
    else:
        fig = ax.figure

    norm = Normalize(vmin=vmin, vmax=vmax)
    colors = cmap(norm(bins["centroid"]))
    floor = np.max(intensity)*fcolorbar

    bar = ax.bar(bins["centroid"], intensity+floor, 
                        width=bins["width"], color=colors,
                   alpha=alpha)
    ax.axhline(floor, color="w", lw=1)
    ax.set_yticks([])
    ax.set_xlim(vmin, vmax)
    if clear_axes:
        clearwhich = ["left","right","top", "bottom"]
        [ax.spines[which].set_visible(False) for which in clearwhich]

    
    return ax



import os
import warnings
import pandas
import numpy as np
IDR_PATH = os.getenv("ZTFIDRPATH", "./dr2")

__all__ = ["get_targets_data",
           "get_host_data",
           "get_localhost_data",           
           "get_autotyping"]
    
# ================== #
#                    #
#   TOP LEVEL        #
#                    #
# ================== #
def get_targets_data():
    """ """
    redshifts = get_redshif_data()[["redshift","redshift_err", "source"]]
    salt2params = get_salt2params()
    coords = get_coords_data()
#    
    # merging
    data_ = pandas.merge(redshifts,salt2params, left_index=True, right_index=True,
                     suffixes=("","_salt"), how="outer")
    data_ = pandas.merge(data_, coords, left_index=True, right_index=True,
                         how="outer")
    # force limit to target to use
    target_list = get_targetlist()
    data_ = data_.loc[target_list["ztfname"]]

    # adding classification
    typing = get_target_typing()
    data_ = data_.join(typing["classification"], how="left")
    return data_

def get_localhost_data(local_nkpc=2, which="mag"):
    """ """
    hostlocal = get_host_local(nkpc=2, which="mag")
    hostcoords = get_host_coords()
    dlr = get_host_mags().xs("global", axis=1)["host_dlr"]
    hostcoords = hostcoords.merge(dlr, left_index=True, right_index=True)
    return hostcoords.merge(hostlocal, left_index=True, right_index=True,
                                how="left")

def get_host_data(local_nkpc=2, which="mag"):
    """ """
    hostmag = get_host_mags().xs("global", axis=1)
    hostcoords = get_host_coords()
    return hostcoords.merge(hostmag, left_index=True, right_index=True,
                                how="left")
                                

# ================== #
#                    #
#   BASICS           #
#                    #
# ================== #
def get_targetlist(load=True, **kwargs):
    """ official list of target to use for dr2 """
    filepath = os.path.join(IDR_PATH, "tables",
                            "ztfdr2_targetlist.csv")
    if not load:
        return filepath
    
    return pandas.read_csv(filepath, **kwargs)

def get_target_typing(load=True, index_col=0, sep=" ",
                          clean=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables/.dataset_creation/sample_def",
                            "ztfdr2_typing.csv")
    if not load:
        return filepath
    
    data = pandas.read_csv(filepath, index_col=index_col, sep=sep, **kwargs)
    if clean:
        data["typing"] = data["typing"].str.replace("[","", regex=False).str.replace("]","", regex=False).str.replace("'","") # clean
        data["typing"] = data["typing"].str.replace("sn ia", "snia").str.replace("not ia", "notia")
        data["typing"] = data["typing"].str.split(" ")
        
        data["ntypings"] = data["ntypings"].str.replace("[","", regex=False).str.replace("]","", regex=False).str.split(" ")
    return data

# Master List
def get_masterlist(load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/sample_def/ztfdr2_masterlist.csv")
    if not load:
        return filepath
    
    data = pandas.read_csv(filepath, **kwargs)
    return data["ztfname"].values.astype(str)

def get_nospectralist(load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/sample_def/ztfdr2_target_nospectra.csv")
    if not load:
        return filepath
    
    data = pandas.read_csv(filepath, names=["ztfname"], **kwargs)
    return data["ztfname"].values.astype(str)


# Redshifts
def get_redshif_data(load=True, index_col=0, full=False, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            "ztfdr2_redshifts.csv")
    if not load:
        return filepath
    
    data = pandas.read_csv(filepath, index_col=index_col, **kwargs)
    data.index.name = 'ztfname'

    if not full: # get only redshift and source
        return data
    
    # or get them all:
    redshift_dir = os.path.join(IDR_PATH, "tables", ".dataset_creation/redshifts")
    redshifts = {}

    
    # Emission lines
    sedm_em = pandas.read_csv(os.path.join(redshift_dir,
                                            "specfile_sedm_emissionlines.csv"),
                                  index_col=0)
    sedm_em = sedm_em.set_index("ztfname")[["sedmredshift","sedmredshift_err"]].copy()#.rename("sedmredshift","sedm_em", axis=1)
    sedm_em.columns = sedm_em.columns.str.replace("sedmredshift","sedm_em")
    
    
    
    nosedm_em = pandas.read_csv(os.path.join(redshift_dir,
                                                    "specfile_nonsedm_emissionlines.csv"),
                                     index_col=0)
    nosedm_em = nosedm_em.set_index("ztfname")[["value","error"]].copy()#.rename("sedmredshift","sedm_em", axis=1)
    nosedm_em.columns = ["nonsedm_em", "nonsedm_em_err"]
    
    # SNID
    snidauto = pandas.read_csv(os.path.join(redshift_dir,
                                                    "ztfdr2_snid_redshifts.csv"),
                                     index_col=0)
    snidauto.columns = snidauto.columns.str.replace("redshift","snidauto")
    
    # Host-Z
    hostz = pandas.read_csv(os.path.join(redshift_dir,
                                                    "ztfdr2_hostz_redshifts.csv"),
                                     index_col=0)
    hostz = hostz[["redshift"]]
    hostz.columns = ["host_cat"]
    hostz["host_cat_err"] = 1e-5


    
    # override
    override = pandas.read_csv(os.path.join(redshift_dir,
                                                    "ztfdr2_override_redshifts.csv"),
                                     index_col=0)
    override.columns = ["override"]

    # merge them
    
    data = data.join(sedm_em, on="ztfname")
    data = data.join(nosedm_em, on="ztfname")
    data = data.join(snidauto, on="ztfname")
    data = data.join(hostz, on="ztfname")
    data = data.join(override, on="ztfname")
    data
    
    return data
    
def get_snidauto_redshift(load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH,"tables",
                                ".dataset_creation/redshifts/ztfdr2_snid_redshifts.csv")
    if not load:
        return filepath
    
    return pandas.read_csv(filepath, index_col=0, **kwargs)

def get_snidauto_specfile_redshift(load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH,"tables",
                                ".dataset_creation/redshifts/snidauto_ianorm.csv")
    if not load:
        return filepath
    
    return pandas.read_csv(filepath, index_col=0, **kwargs)

# Coordinates
def get_coords_data(load=True, index_col=0, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            "ztfdr2_coordinates.csv")
    if not load:
        return filepath
    return pandas.read_csv(filepath, index_col=index_col, **kwargs)

# SALT
def get_salt2params(load=True, default=True, **kwargs):
    """ """
    filename = "ztfdr2_salt2_params.csv" if default else\
      "ztfdr2_salt2_params_phase-15to30_color-0.4to0.8.csv"
    filepath = os.path.join(IDR_PATH, "tables",
                                filename)

    if not load:
        return filepath
    
    return pandas.read_csv(filepath, **kwargs
                          ).rename({"z":"redshift"}, axis=1
                          ).set_index("ztfname")


# ================== #
#                    #
#   HOST             #
#                    #
# ================== #
def get_host_coords(load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/host_prop/host_photometry/ztfdr2_hostcoords.csv")
    if not load:
        return filepath
    
    return pandas.read_csv(filepath, **{**dict(sep=" "),**kwargs}).set_index("ztfname")

def get_host_local(nkpc=2, which="mag", load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            f".dataset_creation/host_prop/host_photometry/ztfdr2_local{nkpc}kpc_{which}.csv")
    if not load:
        return filepath
    if not os.path.isfile(filepath):
        raise IOError(f"no such file {filepath}")

    return pandas.read_csv(filepath, index_col=0, **kwargs)

def get_host_sedfit(nkpc=2, load=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            f".dataset_creation/host_prop/host_properties/ztfdr2_local{nkpc}kpc_sedfit.csv")
    if not load:
        return filepath
    if not os.path.isfile(filepath):
        raise IOError(f"no such file {filepath}")

    return pandas.read_csv(filepath, index_col=0, **kwargs)
    

def get_host_mags(load=True, index_col=0, raw=False, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/host_prop/host_photometry/ztfdr2_hostmags.csv")
    if not load:
        return filepath

    host_alldata = pandas.read_csv(filepath, index_col=index_col, **kwargs)
    if raw:
        return host_alldata

    def _get_split_(which):
        requested = f"{which}_mag"
        if requested not in host_alldata:
            raise ValueError(f"Unknown entry: {requested} (which={which})")

        host_data = host_alldata[requested].astype(str).str.replace("nan", "np.nan").str.replace("inf", "np.nan").apply(eval)
        host_err = host_alldata[requested+"_err"].astype(str).str.replace("nan", "np.nan").str.replace("inf", "np.nan").apply(eval)
        flagna = host_data.isna() + host_err.isna()
        data = pandas.DataFrame(list(host_data[~flagna]), index=host_data[~flagna].index)
        error = pandas.DataFrame(list(host_err[~flagna]), index=host_err[~flagna].index)
        error.columns += "_err"
        return pandas.merge(data, error, left_index=True, right_index=True)

    kpc2 = _get_split_(which="local_2kpc")
    kpc4 = _get_split_(which="local_4kpc")
    host_cat = _get_split_(which="host_cat")
    hglobal = pandas.merge(host_cat,
                          host_alldata[["z","host_dlr"]].rename({"z":"redshift"}, axis=1), 
                          left_index=True, right_index=True,
                               how="outer")
    return pandas.concat([kpc2, kpc4, hglobal], axis=1,
                          keys=["2kpc", "4kpc", "global"])
    
# ================== #
#                    #
#   LIGHTCURVES      #
#                    #
# ================== #
def get_target_lc(target, test_exist=True):
    """ """
    fullpath = os.path.join(IDR_PATH, "lightcurves", f"{target}_LC.csv")
    if test_exist:
        if not os.path.isfile(fullpath):
            warnings.warn(f"No lc file for {target} ; {fullpath}")
            return None

    return fullpath


def get_phase_coverage(load=True, warn=True, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables", "phase_coverage.csv")
    if not load:
        return filepath

    if not os.path.isfile(filepath):
        if warn:
            warnings.warn(
                "No phase_coverage file. build one using ztfidr.Sample().build_phase_coverage(store=True)")
        return None

    return pandas.read_csv(filepath, index_col=0, **kwargs
                           ).reset_index().rename({"index": "name"}, axis=1
                                                  ).set_index(["name", "filter"])["phase"]


# ================== #
#                    #
#   Spectra          #
#                    #
# ================== #
def get_autotyping(load=True, index_col=0, **kwargs):
    """ """
    filepath = os.path.join(IDR_PATH, "tables",
                            ".dataset_creation/sample_def/spec_class/autotyping.csv")
    if not load:
        return filepath
    return pandas.read_csv(filepath, index_col=index_col, **kwargs)


def get_spectra_datafile(contains=None, startswith=None,
                         snidres=False, extension=None, use_dask=False,
                         add_phase=True, data=None):
    """ """
    from glob import glob
    glob_format = "*" if not startswith else f"{startswith}*"
    if snidres and extension is None:
        extension = "_snid.h5"
    elif extension is None:
        extension = ".ascii"

    if contains is not None:
        glob_format += f"{contains}*"
    if extension is not None:
        glob_format += f"{extension}"

    specfiles = glob(os.path.join(IDR_PATH, "spectra", glob_format))
    datafile = pandas.DataFrame(specfiles, columns=["fullpath"])
    datafile["basename"] = datafile["fullpath"].str.split(pat="/", expand=True).iloc[:, -1]
    
    specfile = pandas.concat([datafile, parse_filename(datafile["basename"], snidres=snidres)], axis=1)
    
    if add_phase:
        from astropy.time import Time
        if data is None:
            from .sample import get_sample
            data = get_sample().data
            
        specfile["dateiso"] = Time(np.asarray(specfile["date"].apply(lambda x: f"{x[:4]}-{x[4:6]}-{x[6:]}"), dtype=str), format="iso").mjd
        specfile = specfile.join(data["t0"], on="ztfname")
        specfile["phase"] = specfile.pop("dateiso")-specfile.pop("t0")
        
    return specfile


def parse_filename(file_s, snidres=False):
    """ file or list of files.
    Returns
    -------
    Serie if single file, DataFrame otherwise
    """

    index = ["ztfname", "date", "telescope", "version"]
    fdata = []
    for file_ in np.atleast_1d(file_s):
        file_ = os.path.basename(file_).split(".ascii")[0]
        if snidres:
            #print(file_)
            name, date, *telescope, origin, snid_ = file_.split("_")
        else:
            try:
                name, date, *telescope, origin = file_.split("_")
            except:
                print(f"failed parsing filename for {file_}")
                continue

        telescope = "_".join(telescope)
        fdata.append([name, date, telescope, origin])

    if len(fdata) == 1:
        return pandas.Series(fdata[0], index=index)
    return pandas.DataFrame(fdata, columns=index)



""" Interact with the typingapp """

from . import io
import os
import pandas
import numpy as np


DB_PATH = os.path.join(io.IDR_PATH, "typingapp.db")

# typing app data
import sqlite3
con = sqlite3.connect(DB_PATH)


SORT_FAVORED = ["not ia","ib/c","ii","gal", "other",
                "sn ia","ia-norm",
                "ia-91bg", "ia-91t", "ia-other"]

def get_typing(typings_values, 
               sort_favored=SORT_FAVORED):
    """ """
    typings, values = typings_values
    typings[values==values.max()]
    if len(typings)  == 1:
        return typings[0]
    
    ltypings = list(typings)
    ltypings.sort(key = lambda i: sort_favored.index(i))
    return ltypings[0]

def get_specs_to_rm():
    """ get the list of spectra flaged as 'to remove' in the typingapp """
    report = Reports()
    rmspec_serie = report.data[(report.data["value"].str.startswith("spec:rm")) &
                               (~report.data["value"].str.contains("none_left"))]
    rmspecs = np.asarray(rmspec_serie["value"].str.replace("spec:rm:","").values, dtype=str)
    return rmspecs

def parse_classification(line, 
                         min_review=2,
                         min_autotyping=3, 
                         min_generic_typing=3,
                         unclear_limit=0.5,
                         to_unclear_limit=0.75,
                         min_classifications_to_unclear=6,
                         verbose=False, add_origin=False, skipsave=False):
    """ """
    classification = "confusing"

    if verbose:
        print(line)
        print(line["typing"])
        print("unclear" in line["typing"])
        
    # not enough review, pass
    if line.nreviews <= min_review: # not enough
        if verbose:
            print("not enough classification")
        classification = "None"
        level = 0
        
    # Level 1
    # easy, all the same 
    elif line.ntypes == 1 and line.ntypings[0] >= min_autotyping:
        if verbose:
            print("all the same")
        classification =  line.typing[0]
        level = 1
    # special save for ia-norm cases.

    # Level 2    
    # enough for generic, let's see if it's sufficient.        
    elif line.nreviews >= min_generic_typing:
        if verbose:
            print("enough for generic review")
            
        # go back to the easy-case
        if "ia-norm" in line.typing and np.all([k in ["ia-norm", "sn ia", "unclear"] for k in line.typing]) \
          and line.ntypings[ list(line.typing).index("ia-norm") ] >= min_autotyping:
            classification = "ia-norm"
            
        # It looks like a Ia ir Ia-norm
        elif np.all([k in ["ia-norm", "sn ia"] for k in line.typing]):
            classification = "ia(-norm)"
            
        # Typing ok, Sub-tyiping not            
        elif np.all(['ia' in k and "not ia" not in k for k in line.typing]):
            classification = "ia(-?)"
            
        # all non-ia ?
        elif np.all([k in Classifications.NONIA_KEY for k in line.typing]):
            classification = "nonia"
            
        else:
            classification = "confusing"

        level = 2
    else:
        level = -1


    # Level 3            
    # saving unclears if possible
    # try to save remaining confusing cases
    if classification == "ia(-?)" and not skipsave:
        # remove generic cases
        new_line = clear_line(line.copy(), ["unclear", "sn ia"])
        classification, _ = parse_classification(new_line, skipsave=True)
        # Could not be saved:
        if classification in ["None", "unclear", "confusing","ia(-?)"]:
            classification = "ia(-?)"
        else:
            if verbose:
                print(f"ia(-?) classification became {classification}")
            
    # Level 3            
    # saving unclears if possible
    # try to save remaining confusing cases
    elif classification == "confusing" and "unclear" in line["typing"]:
        
        if verbose:
            print("trying to save confusing cases")
            print("unclear" in line["typing"])
            
        info = dict(zip(line["typing"],line["ntypings"]))
        n_unclears = info["unclear"] 
        n_not = np.sum([v for k,v in info.items() if k !="unclear"])
        n_classifications = n_unclears + n_not
        f_unclear = n_unclears / (n_classifications)
        f_not = 1-f_unclear
        if verbose:
            print(f"fraction of confusing: {f_unclear}" )
            
        # more than 4 classification agreed when unclear discarded.
        if f_unclear <= unclear_limit and n_not >=min_generic_typing:
            if verbose:
                print(f"saving this case" )

            new_line = clear_line(line.copy())
            classification, _ = parse_classification(new_line)

        # At least 6 classification and more than 4 (if to_unclear_limit==0.75) are 'unclear'
        elif f_unclear >= to_unclear_limit and n_classifications >= min_classifications_to_unclear:
            if verbose:
                print(f"saving this case as 'unclear' ")
            classification = "unclear"
        elif n_not <= min_review:
            if verbose:
                print(f"saving this case as 'None' ")
            classification = "None"
            
        if verbose:
            print(f"cannot save this case ( {f_unclear} > {to_unclear_limit} & {n_classifications}>{n_classifications}")

        level = 3

    # Back to saving the rest:
    if classification == "None":
        if line.nreviews >= 2:
            # 2 unclears or more
            if np.all([k in ["unclear"] for k in line.typing]):
                classification = "unclear"
                if verbose:
                    print(f"saving None to {classification}")
                
            else:
                line = clear_line(line, "unclear")
                # still have 2 or more entries
                if line.nreviews >= 2:
                    # No ?
                    # let's remove the unclear then, and let's see:
                    line = clear_line(line, "unclear")
                    # All the same
                    if line.ntypes == 1:
                        classification =  line.typing[0]
                        if verbose:
                            print(f"saving None to {classification}")
                
                    elif np.all([k in ["ia-norm", "sn ia"] for k in line.typing]):
                        classification = "ia(-norm)"
                        if verbose:                        
                            print(f"saving None to {classification}")
                
                    elif np.all([k in ["unclear"] for k in line.typing]):
                        classification = "unclear"
                        if verbose:                        
                            print(f"saving None to {classification}")
                
                    elif np.all(['ia' in k and "not ia" not in k for k in line.typing]):
                        classification = "ia(-?)"
                        if verbose:
                            print(f"saving None to {classification}")
                
    return classification, level


def clear_line(line, rm_value="unclear"):
    """ remove the 'unclear' component from the given typing row """
    typing = list(line.typing)
#    if "unclear" not in typing:
#        return line # nothing to do
    
    ntypings = list(line.ntypings)
    for value in np.atleast_1d(rm_value):
        if value not in typing:
            continue
        index_unclear = typing.index(value)
        typing.pop(index_unclear)
        ntypings.pop(index_unclear)
        
    return pandas.Series({"typing": typing,
                          "ntypings": ntypings,
                          "ntypes": len(ntypings),
                          "nreviews": np.sum(ntypings)})


class _DBApp_():
    _DB_NAME = None

    def __init__(self):
        """ """
        self.load()
    
    @classmethod
    def data_from_db(cls, sql=None):
        """ """
        basequery = f"SELECT * FROM {cls._DB_NAME}"
        if sql is not None:
            basequery +=f" {sql}"
            
        return pandas.read_sql_query(basequery, con)
        
        
    def load(self):
        """ """
        self._data = self.data_from_db()
    
    # =============== #
    #   Properties    #
    # =============== #
    @property
    def data(self):
        """ """
        return self._data

# =================== #
#                     #
#  USERS              #
#                     #
# =================== #    
class _Users_( _DBApp_ ):
    _DB_NAME = "Users"
    
    def load(self):
        """ """
        self._data = self.data_from_db()[["id","username","name","email","date_added"]]

# =================== #
#                     #
#  CLASSIFICATIONS    #
#                     #
# =================== #
class Reports( _DBApp_ ):
    _DB_NAME = "Classifications"
    
    def load(self):
        """ """
        self._data = self.data_from_db(sql="where kind='report'")

    @classmethod
    def get_arbiters(cls, prefix="arbiter_"):
        """ """
        report = cls()
        dataarbiter = report.data[ report.data["value"].str.startswith("arbiter") ].copy()
        dataarbiter[f"{prefix}classification"] = dataarbiter["value"].str.replace("arbiter:","")
        a = dataarbiter["arbiter_classification"
                       ].str.replace("^ia","ia-", regex=True
                            ).replace("gal","nonia",regex=False
                            ).replace("snia","sn ia",regex=False
                            ).replace("notia","nonia",regex=False
                            ).replace("ibc","ib/c",regex=False) # self consistency
        a[a=="ia-"] = "sn ia"
        dataarbiter["arbiter_classification"] = a.copy()
        
        return dataarbiter
        
    
# =================== #
#                     #
#  CLASSIFICATIONS    #
#                     #
# =================== #
class Classifications( _DBApp_ ):
    _DB_NAME = "Classifications"
    
    MASTER_USER = ["jesper", "Joel Johansson", "Kate Maguire",
                   "Mathew Smith", "Umut", "Georgios Dimitriadis"]

    NONIA_KEY = ["not ia", "ii","ib/c","gal","other", "nonia"]
    IA_PEC_KEY = ["ia-91bg","ia-91t","ia-other"]
    CLASSIFIED_KEY = ["ia-norm", "ia-pec", "ia(-?)", "sn ia", "ia(-norm)"]
    

    def __init__(self, in_targetlist=None):
        """ """
        _ = super().__init__()
        
        if in_targetlist is not None:
            self._data = self._data[self._data["target_name"].isin(in_targetlist)]
            self._targetlist = self.data["target_name"].unique()
        else:
            self._targetlist = io.get_masterlist()
            
    
    def load(self):
        """ """
        self._data = self.data_from_db(sql="where kind='typing'")
        self._data["value"] = self._data["value"].str.strip().str.lower()

    def load_sample(self):
        """ """
        from . import get_sample
        self._sample = get_sample()

    def get_typing(self, **kwargs):
        """ """
        if "classification" not in self.data:
            self.load_classification(**kwargs)
            
        data = self.get_classification_df()
        classification = self.data.groupby("target_name").first()["classification"]
        data = data.join(classification)
        return data
    
    def store_typing(self, **kwargs):
        """ """
        data = self.get_typing(**kwargs)
        return data.to_csv( io.get_target_typing(False), sep=" ")

    def load_classification(self, **kwargs):
        """ """
        classifications = self.get_classification(**kwargs)
        self._data = self.data.join(classifications, on="target_name", how="outer")
        
    def get_classification(self, min_review=2,
                                  min_autotyping=3,
                                  min_generic_typing=3,
                                # Master keys
                                  min_review_master=1,
                                  min_autotyping_master=2,
                                  min_generic_typing_master=3):
        """ """
        # All auto | default
        auto_data = self._get_classification(prefix="auto_",
                                                 min_review=min_review, min_autotyping=min_autotyping,
                                                 min_generic_typing=min_generic_typing)

        # Masters
        master = self.get_masterclassification(incl_unclear=False) # do not consider unclear's input
        master_data = master._get_classification(prefix="master_",
                                                      min_review=min_review_master, min_autotyping=min_autotyping_master,
                                                      min_generic_typing=min_generic_typing_master)
        # do not consider master's None or confusing. Only final classification

        # Arbiter
        arbiter_data = Reports.get_arbiters(prefix="arbiter_").groupby("target_name").last()[["arbiter_classification"]] # [[ to get a dataframe
        # - force in data list
        arbiter_data = arbiter_data.loc[ arbiter_data.index.isin(self.target_list) ]

        #
        # Now let's join
        #
        # - Default
        cdata = pandas.DataFrame(auto_data["auto_classification"].values, index=auto_data.index, columns=["classification"])
        cdata["class_origin"] = "auto"
        cdata = cdata.join(auto_data)
        
        # - add master
        # do not consider bad cases
        master_data = master_data[~master_data["master_classification"].isin(["None","confusing", np.NaN])]
        
        cdata.loc[master_data.index, "classification"] = master_data["master_classification"]
        cdata.loc[master_data.index, "class_origin"] = "master"
        cdata = cdata.join(master_data)
        
        # - add arbiter
        cdata.loc[arbiter_data.index, "classification"] = arbiter_data["arbiter_classification"]
        cdata.loc[arbiter_data.index, "class_origin"] = "arbiter"
        cdata = cdata.join(arbiter_data)

        return cdata
        
        
    def _get_classification(self, prefix="", reindex=True, **kwargs):
        # Normal
        class_df = self.get_classification_df()
        d_ = pandas.DataFrame(class_df.apply(parse_classification, axis=1, **kwargs).to_list(),
                                columns=[f"{prefix}classification",f"{prefix}clevel"], index=class_df.index)
        return d_.reindex(list(self.target_list))
    # -------- #
    # GETTER   #
    # -------- #
    def get_masterclassification(self, incl_unclear=True):
        """ """
        this = self.__class__()
        MASTER_ID = _Users_().data.set_index("name").loc[self.MASTER_USER]["id"].values
        this._data = self.data[self.data["user_id"].isin(MASTER_ID)]
        if not incl_unclear:
            this._data = this._data[~(this._data["value"] == "unclear")]
            
        return this

    def get_classification_stats(self, based_on="classification", **kwargs):
        """ Get the distribution of classifications made. 

        Parameters
        ----------
        based_on: str
            classification key to used (from self.data); could be:
            - classification
            - master_classification
            - auto_classification

        **kwargs goes to get_data()

        Returns
        -------
        pandas.Series
            groupby().size() serie.
        """
        if "classification" not in self.data:
            self.load_classification()

        data = self.get_data(**kwargs)
        return data.groupby("target_name").first().groupby(based_on).size()
    
    
    def get_data(self, incl_unclear=True, min_classification=None,
                     types=None, 
                     goodcoverage=None, redshift_range=None,
                     x1_range=None, c_range=None, **kwargs):
        """ 

        incl_unclear: [bool] -optional-
            should the unclear typing be included ?

        min_classification: [None or int] -optional-
            minimum number of typing for this target. 
            'unclear' are not counted as classifications
            (this uses self.get_ntyped()) 


        goodcoverage: [None or bool] -optional-
            # sample.get_data() options
            should we only use target with good lightcurve sampling ?
            - None: no selection
            - True: only good sampling targets
            - False: only bad sample targets
            
        redshift_range, x1_range, c_range: [None or [min, max] ] -optional-
            # sample.get_data() options
            range to limit redshift, x1 or c, respectively
        
        **kwargs goes to self.sample.get_data()

        Returns
        -------
        DataFrame
        """
        data = self.data.copy()
        if not incl_unclear:
            data = data[data["value"] != "unclear"]
            
        if min_classification is not None and min_classification>0:
            ntyped = self.get_ntyped(incl_unclear=False)
            data = data[data["target_name"].isin(ntyped[ntyped>=min_classification].index)]
            
        if types is not None:
            data = data[data["value"].isin(np.atleast_1d(types))]
            
        # Sample
        _locals = locals()
        prop_sample = kwargs
        for k in ["goodcoverage", "redshift_range", "x1_range", "c_range"]:
            if _locals[k] is not None:
                prop_sample[k] = _locals[k]
                
        if len(prop_sample) >0:
            names_to_keep = self.sample.get_data(**prop_sample).index
            data = data[data["target_name"].isin(names_to_keep)]
            
        return data

    def get_ntyped(self, incl_unclear=False):
        """ series of the n"""
        return self.get_data(incl_unclear=incl_unclear).groupby("target_name").size().sort_values()
    
    def get_target_data(self, name):
        """ """
        return self.data.query(f"target_name == '{name}'")
    
    def get_of_types(self, types):
        """ get entry associated to the given type (lower case) """
        return 
                         
    def get_nonia(self, incl_unclear=False):
        """ """
        list_to_get = self.NONIA_KEY.copy()
        if incl_unclear:
            list_to_get+="unclear"
            
        return self.data[self.data["value"].isin(list_to_get)]

    def get_classification_df(self):
        """ """
        data = self.get_data(incl_unclear=True)
        typings = data.groupby(["target_name"])["value"].apply(np.unique, return_counts=True)
        types =  pandas.DataFrame(typings.to_list(), columns=["typing", "ntypings"],
                                index=typings.index)#.explode(["typing","ntypings"])
        types["ntypes"] = types["ntypings"].apply(len)
        types["nreviews"] = types["ntypings"].apply(np.sum)
        return types
    
    # -------- #
    # PLOTTER  #
    # -------- #
    def show_classification_stats(self, explode = 0.1,
                                  based_on = "classification", 
                                  redshift_range = None, dataprop = {},
                                  startangle_main=90, startangle_sub=0,
                                  **kwargs):
        """ """
        import matplotlib.pyplot as plt
        
        prop_general = {**dict(wedgeprops=dict(width=1, edgecolor='0.7', linewidth=0.2)), 
                **kwargs}


        class_stat = self.get_classification_stats( based_on, redshift_range=redshift_range,
                                                    **dataprop)

        class_stat["ia-pec"] = class_stat[self.IA_PEC_KEY].sum()
        class_stat["Type Ia"] = class_stat[self.CLASSIFIED_KEY].sum()
        class_stat["non ia"] = class_stat[[k for k in self.NONIA_KEY if k in class_stat]].sum()
        print(class_stat["non ia"])
        # showing plots


        fig = plt.figure(figsize=[9,4])
        ax_main = fig.add_axes([0.05,0.15,0.4,0.8])
        ax_sub = fig.add_axes([0.52,0.15,0.4,0.8])

        # -> Main plot
        to_show = ["None","confusing", "unclear", "Type Ia", "non ia"]
        data_show = class_stat[to_show]
        if type(explode) == float:
            explode = np.full_like(data_show.values, explode, dtype="float")

        colors = ["w","0.8", plt.cm.bone(0.3),  plt.cm.bone(0.7),
                  "tab:brown"]
        prop = {**dict(colors=colors, explode=explode, autopct='%.0f%%', 
                       startangle=startangle_main), **prop_general}

        # PLOTTING
        _ = ax_main.pie(data_show.values, labels=data_show.index, **prop)

        ax_main.text(0.5, -0.15, f"{np.sum(data_show.values)} Supernovae", 
                     transform=ax_main.transAxes,
                     fontsize="large", ha="center", va="bottom")

        # -> sub plot
        to_show = self.CLASSIFIED_KEY
        data_show = class_stat[to_show]
        if type(explode) == float:
            explode = np.full_like(data_show.values, explode, dtype="float")

        colors = ["tab:blue", "tab:orange", "wheat", "lavender", "lightsteelblue"]
        prop = {**dict(colors=colors, explode=explode, autopct='%.0f%%', 
                       startangle=startangle_sub), **prop_general}

        # PLOTTING
        _ = ax_sub.pie(data_show.values, labels=data_show.index, **prop)

        ax_sub.text(0.5, -0.15, f"{np.sum(data_show.values)} Type Ia Supernovae", 
                    transform=ax_sub.transAxes,
                    fontsize="large", ha="center", va="bottom")

        return fig

    
    def show_classifications(self, ax=None, fig=None, 
                                 classifications=None, per_target=True):
        """ """
        if classifications is not None:
            ndatas = classifications.value_counts().sort_values()
        elif per_target:
            ndatas = self.get_classification().value_counts().sort_values()
        else:
            ndatas = self.data.groupby(["value"]).size().sort_values()
        
        if ax is None:
            if fig is None:
                import matplotlib.pyplot as mpl
                fig = mpl.figure(figsize=[7,4])
            ax = fig.add_subplot(111)
        else:
            fig = ax.figure
        
        hbars = ax.barh(ndatas.index, ndatas, tick_label=ndatas.index)
        ax.tick_params("y", labelsize="small", color="C0", length=0)

        ax.bar_label(hbars, fmt='%d', color="C0", padding=1, 
                     fontsize="x-small")

        clearwhich = ["bottom","right","top", "left"] # "bottom"
        [ax.spines[which].set_visible(False) for which in clearwhich]
        ax.set_xticks([])
        return fig
        
    # ============== #
    #  Properties    #
    # ============== #
    @property
    def sample(self):
        """ """
        if not hasattr(self, "_sample"):
            self.load_sample()
            
        return self._sample

    @property
    def target_list(self):
        """ """
        return self._targetlist
    
    @property
    def types(self):
        """ list of all possible types"""
        return self.data["value"].unique()

    @property
    def ia_types(self):
        """ list of all possible types"""
        alltypes = self.types
        return alltypes[["ia" in _ and _ !="not ia" for _ in alltypes ]]



import numpy as np
import matplotlib.pyplot as plt

from . import utils

def show_hubble_standardisation(sample, fig=None,
                                param="c", vmin=-0.2, vmax=0.3, cmap="coolwarm", 
                                clear_axes=False):
    """ """
    from astropy.cosmology import Planck18
    data = sample.get_data(x1_range=[-4,4], c_range=[-0.3,0.8], goodcoverage=True)
    data = data[data["classification"].isin(['ia(-norm)','ia-norm'])]
    
    if fig is None:
        fig = plt.figure(figsize=[6,4])
    
    ax = fig.add_axes([0.1,0.15,0.8,0.75])
    cax = fig.add_axes([0.4,0.3,0.4,0.25])

    mag = -2.5 * np.log10(data["x0"])
    sc = ax.scatter(data["redshift"], 
                    mag+19*1.58, c=data[param], 
                    cmap=cmap,
                    s=10, vmin=vmin, vmax=vmax)


    _ = utils.hist_colorbar(data[param], ax=cax, cmap=cmap, fcolorbar=0.1,
                           vmin=vmin, vmax=vmax)

    xx = np.linspace(0.002,0.23,1000)
    ax.plot(xx, Planck18.distmod(xx), color="k", lw=2)

    ax.set_ylabel("distance modulus + cst", fontsize="large")
    ax.set_xlabel("redshift", fontsize="large")
    cax.set_xlabel(f"{param}", fontsize="small", color="0.5")
    cax.tick_params(axis="x", labelsize="small", 
                   labelcolor="0.5", color="0.5")
    zref = 0.204
    ax.text(zref, Planck18.distmod(zref).value, "Planck H(z)", va="center", ha="center", 
           fontsize="x-small", rotation=9, weight="bold",
            bbox={"facecolor":"w", "edgecolor":"None"})
    
    if clear_axes:
        clearwhich = ["bottom","right","top","left"] # "bottom"    
        [ax.spines[which].set_visible(False) for which in clearwhich]
        ax.tick_params(labelsize="medium")
#    ax.set_xticks([])

    return fig


def show_typingdistribution(sample, ax=None, fig=None):
    """ """
    dist_typing = sample.data.groupby("classification").size()
    ia_norm = dist_typing[["ia-norm","ia(-norm)"]].sum()
    ia_pec = dist_typing[["ia-91t","ia-91bg","ia-other"]].sum()
    ia = dist_typing[["sn ia"]].sum()
    rest = dist_typing.sum() - (ia_norm+ia_pec+ia)
    if ax is None:
        if fig is None:
            fig = plt.figure(figsize=[7,3])
        ax = fig.add_subplot(111)
    else:
        fig = ax.figure

    ax.barh(1, ia_norm, facecolor="C0", edgecolor="None")
    ax.barh(0, ia, facecolor="0.7", edgecolor="None")
    ax.barh(-1, ia_pec, facecolor="C1", edgecolor="None")
    ax.barh(-2, rest, facecolor="None", edgecolor="k")
        
    ax.text(ia_norm+30,1, f'{ia_norm}', fontsize="large",
            color="C0", va="center", ha="left", weight="bold")
    ax.text(ia+30,0, f'{ia}', fontsize="large",
            color="0.7", va="center", ha="left", weight="bold")
    ax.text(ia_pec+30,-1, f'{ia_pec}', fontsize="large",
            color="C1", va="center", ha="left", weight="bold")

    ax.text(rest+30,-2, f'{rest}', fontsize="large",
            color="k", va="center", ha="left")

    clearwhich = ["bottom","right","top",] # "bottom"
    [ax.spines[which].set_visible(False) for which in clearwhich]
    ax.set_xticks([])
    ax.set_yticks([1,0,-1,-2], ["SN Ia\nnorm","SN Ia","SN Ia\npeculiar", "Unclear"])

    return fig



from scipy.stats import median_abs_deviation
import numpy as np
import pandas

def estimate_typing(snidresult,  weight_by="rlap",
                    topn=30, rlap_range=[5,1000], 
                    grad="good", default=None):
    """ """
    
def estimate_redshift(snidresult, weight_by="rlap",
                              topn=30, typing=None, rlap_range=[5,1000], 
                              grad="good", default=[np.NaN, np.NaN]):
    """ """
    
    d = snidresult.copy()
    if topn is not None:
        d = d[d["no."].astype("int")<30]
    if typing is not None:
        d = d[d["type"].isin(np.atleast_1d(typing))]
    if rlap_range is not None:
        d = d[d["rlap"].between(*rlap_range)]
        
    if len(d) == 0:
        return default
    if len(d) == 1:
        return d["z"][0], np.NaN
    
    redshift = np.average(d["z"], weights=d[weight_by])
    dredshift = median_abs_deviation(d["z"])
    return redshift, dredshift
